(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Sep 2017
 *)
(* begin show *)

(** **** Global imports **)
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.
From Coq Require Import Bool.

(** **** Local  imports **)
From Padec Require Import SetoidUtil.
From Padec Require Import ZUtils.
From Padec Require Import OptionUtil.
From Padec Require Import ListUtils.
From Padec Require Count.

(* end show *)
Open Scope signature_scope.
Set Implicit Arguments.

(** * Decidability Properties over a finite type stored into a list *)
Section All_In.

  Context {A: Type} `{SA: Setoid A}.
  Variable (all_A: list A)
           (all_A_prop: forall a, InA equiv a all_A).

  Lemma All_In_dec_forall_exists:
    forall (P: A -> Prop) (P_compat: Proper fequiv P)
           (P_dec: forall a, {P a} + {~P a}),
      {forall a, ~P a} + {exists a, P a}.
  Proof.
    intros P P_compat P_dec.
    pose (P_b := (fun a => if P_dec a then true else false)).
    assert (P_b_compat: Proper fequiv P_b).
    { intros a1 a2 ea; unfold P_b;
        destruct (P_dec a1); destruct (P_dec a2); simpl; auto.
      contradict n; apply P_compat with a1; [symmetry|]; auto.
      contradict n; apply P_compat with a2; auto.
    }
    assert (F: existsb P_b all_A = false <-> (forall a, ~P a)).
    {
      rewrite existsb_forallb with (notf := fun a => negb (P_b a)).
      rewrite negb_false_iff.
      rewrite (forallb_forall_InA (A:=A)).
      apply all_compat; intro a; intuition.
      specialize (H (all_A_prop a)).
      rewrite negb_true_iff in H.
      contradict H; rewrite not_false_iff_true.
      unfold P_b; destruct (P_dec a); congruence.
      rewrite negb_true_iff.
      unfold P_b; destruct (P_dec a); congruence.
      intros a1 a2 ea; apply negb_sym; rewrite negb_involutive.
      symmetry; apply P_b_compat, ea.
      intros a; rewrite negb_involutive; auto.
    }
    assert (T: existsb P_b all_A = true <-> (exists a, P a)).
    {
      rewrite existsb_exists_InA.
      apply ex_compat; intro a; intuition.
      unfold P_b in H3; destruct (P_dec a); congruence.
      unfold P_b; destruct (P_dec a); congruence.
      apply P_b_compat.
    }
    case_eq (existsb P_b all_A); intro h; [right|left]; tauto.
  Qed.

  Lemma All_In_dec_forall_notforall:
    forall (P: A -> Prop) (P_dec: forall a, {P a} + {~P a}),
      {Proper fequiv P -> forall a, P a} + {~(forall a, P a)}.
  Proof.
    intros P P_dec.
    destruct (Bool.bool_dec
                (forallb (fun a => if P_dec a then true else false)
                         all_A) true) as [T|T].
    { left; intros P_compat.
      rewrite forallb_forall_InA in T.
      intros a.
      specialize (all_A_prop a); specialize (T a all_A_prop).
      destruct (P_dec a); [auto | congruence].
      intros a1 a2 ea; specialize (P_compat _ _ ea);
        destruct (P_dec a1); destruct (P_dec a2); simpl;
          simpl in P_compat; tauto.
    }
    { right; intros F; contradict T.
      rewrite forallb_forall.
      intros a _; specialize (F a); destruct (P_dec a); tauto.
    }
  Qed.

  Lemma All_In_dec_forall_ex0:
    forall (DA: Decider (equiv (A := A))) (l: list A),
    forall (P: A -> Prop) (P_compat: Proper fequiv P)
           (P_dec: forall a, {InA equiv a l -> P a} + {~P a}),
      {a | InA equiv a l /\ P a} +
      {forall a, InA equiv a l -> ~P a}.
  Proof.
    intros DA l P P_compat P_dec.
    clear all_A_prop all_A.
    induction l as [| a l ih].
    { right; intros a h; rewrite InA_nil in h; tauto. }
    { case (P_dec a); [left; exists a; auto|].
      intros np.
      case ih.
      { intros b; destruct (P_dec b) as [hb|hb]; [|right; auto].
        case (DA a b); intros ha.
        right; contradict np; revert np; apply P_compat; auto.
        left; intros h; apply hb; rewrite InA_cons; right; auto.
      }
      { intros [b [hb hp]]; left; exists b; split; auto. }
      { intros hp; right; intros b hb; rewrite InA_cons in hb.
        destruct hb as [hb|hb].
        contradict np; revert np; apply P_compat; symmetry; auto.
        apply hp; auto.
      }
    }
  Qed.

  Lemma All_In_dec_forall_ex:
    forall (P: A -> Prop) (P_compat: Proper fequiv P)
           (P_dec: forall a, {P a} + {~P a}),
      {a | P a} + {forall a, ~P a}.
  Proof.
    intros P; intros P_compat P_dec.
    cut ({a | P a} + {forall a, InA equiv a all_A -> ~P a}).
    { intros [ok|ok]; [left|right]; auto. }
    clear all_A_prop.
    induction all_A as [| a all_A_ ih].
    { right; intros a h; rewrite InA_nil in h; tauto. }
    { case (P_dec a); [left; exists a; auto|].
      intros np.
      case ih; [left | right]; auto.
      intros aa ina; rewrite InA_cons in ina; destruct ina as [ina|ina].
      - rewrite ina; auto.
      - apply n; auto.
    }
  Qed.

  Lemma All_In_dec_forall_ex':
    forall (P: A -> Prop) (P_compat: Proper fequiv P)
           (P_dec: forall a, {P a} + {~P a}),
      {a | ~ P a} + {forall a, P a}.
  Proof.
    intros.
    assert (np: Proper pequiv (fun a => ~ P a)).
    { intros a b ab; rewrite ab; simpl; tauto. }
    assert (npd: forall a : A, {~P a} + { ~~P a}).
    { intros a; destruct (P_dec a); [right|left]; auto. }
    destruct(All_In_dec_forall_ex np npd); [left|right]; auto.
    intros a; destruct (P_dec a); auto.
    specialize (n a); tauto.
  Qed.
  
  (** ** Emptiness Test Enum:
   Proves emptyness test (i.e. the fact that a function
   always returns None) using
   the assumption that every element of type A is stored in a list
   (hence finite enumerable).
   *)
  Lemma emptiness_by_enumeration:
    forall B (eqB: relation B) (f: A -> option B),
      {Proper (equiv ==> eqoptionA eqB) f ->
       (equiv ==> eqoptionA eqB) f (fun _ => None)}
      +
      {~(equiv ==> eqoptionA eqB) f (fun _ => None)}.
  Proof.
    intros B eqB f.
    destruct (Bool.bool_dec (forallb (fun a => if f a then false else true)
                                     all_A) true) as [T|T].
    { left; intros f_compat.
      intros a _ _.
      specialize (all_A_prop a); rewrite InA_alt in all_A_prop.
      destruct all_A_prop as [a' [ea ina]].
      rewrite forallb_forall in T; specialize (T _ ina).
      specialize (f_compat _ _ ea).
      destruct (f a); destruct (f a'); try tauto; try congruence.
    }
    { right; intros F; contradict T.
      rewrite forallb_forall; intros a _.
      specialize (F _ _ (Equivalence_Reflexive a)).
      destruct (f a); auto.
    }
  Qed.

  (** as above, emptiness is decidable
    + when f is not None everywhere, we can pick some node [n] such that
    [(f n)] is not [None]. *)
  Lemma emptiness_by_enumeration_bis:
    forall B (eqB: relation B)
           (f: A -> option B) (f_compat: Proper (equiv ==> eqoptionA eqB) f),
      { n | exists s, f n = Some s }
      +
      {(equiv ==> eqoptionA eqB) f (fun _ => None)}.
  Proof.
    intros B eqB f f_compat.
    case (emptiness_by_enumeration eqB f); intro empty_test;
      [right; apply empty_test; auto |].
    left.
    assert (em: (forall a, InA equiv a all_A ->
                           eqoptionA eqB (f a) None) -> False).
    { intro h.
      apply empty_test.
      intros a _ _; apply h; auto.
    }
    clear all_A_prop empty_test.
    induction all_A as [| a l ih].
    - contradict em.
      intros a F; apply InA_nil in F; tauto.
    - case_eq (f a); [intros s hs | intro hn].
      * exists a; exists s; auto.
      * apply ih; intros F; apply em.
        intros b inb; apply InA_cons in inb; destruct inb as [eb | inb].
        rewrite <- hn; apply f_compat, eb.
        apply F; auto.
  Qed.

End All_In.

Section All_In_Count.

  Context {A: Type} `{SA: Setoid A} {A_dec: Decider (equiv (A := A))}.
  Variable (all_A: list A)
           (all_A_prop: forall a, InA equiv a all_A).

  Lemma All_In_exists_Card_T:
    forall (P: A -> Prop) (P_compat: Proper pequiv P)
           (P_dec: forall x, {P x}+{~ P x}),
    { NP | Count.Num_Card Count.Same { x | P x } NP}.
  Proof.
    intros P P_compat P_clas.
    eapply Count.Classical_subset_Num_CardT; auto.
  Qed.

  Lemma All_In_exists_Card:
    forall (P: A -> Prop) (P_compat: Proper pequiv P)
           (P_clas: forall x, P x \/ ~ P x),
    exists NP, Count.Num_Card Count.Same { x | P x } NP.
  Proof.
    intros P P_compat P_clas.
    eapply Count.Classical_subset_Num_Card; auto.
    intros n1 n2; destruct (A_dec n1 n2); [left|right]; auto.
  Qed.

End All_In_Count.

Section All_In_2.

  Context {A: Type} `{SA: Setoid A}.
  Context {B: Type} `{SB: Setoid B}.

  Fixpoint build_A_ListB (a: A) (l: list B): list (A*B) :=
    match l with
    | nil => nil
    | b :: lb => (a, b) :: (build_A_ListB a lb)
    end.

  Lemma all_build_A_ListB:
    forall a lb b aa,
      InA equiv b lb -> aa == a -> InA equiv (aa, b) (build_A_ListB a lb).
  Proof.
    induction lb as [| b lb ih].
    - intros b aa F; rewrite InA_nil in F; tauto.
    - intros bb aa inbb eaa; simpl; rewrite InA_cons in *;
        destruct inbb as [ebb|inbb]; simpl.
      left; split; simpl; auto.
      right; apply ih; auto.
  Qed.

  Fixpoint build_ListA_ListB (la: list A) (lb: list B): list (A*B) :=
    match la with
    | nil => nil
    | a :: la => (build_A_ListB a lb) ++ (build_ListA_ListB la lb)
    end.

  Lemma all_build_ListA_ListB:
    forall la lb a b,
      InA equiv a la -> InA equiv b lb ->
      InA equiv (a, b) (build_ListA_ListB la lb).
  Proof.
    induction la as [| a la ih].
    - intros lb a b F; rewrite InA_nil in F; tauto.
    - intros lb aa b inaa inb; simpl; rewrite InA_cons in *;
        rewrite InA_app_iff; destruct inaa as [eaa|inaa].
      left; apply all_build_A_ListB; auto.
      right; apply ih; auto.
  Qed.

  Variable (all_A: list A) (all_A_prop: forall a, InA equiv a all_A).
  Variable (all_B: list B) (all_B_prop: forall a, InA equiv a all_B).
  Definition all_AB: list (A*B) := build_ListA_ListB all_A all_B.

  Lemma all_AB_prop: forall ab, InA equiv ab all_AB.
  Proof.
    intros [a b]; apply all_build_ListA_ListB; auto.
  Qed.


  Lemma All_In_dec_forall_exists_2:
    forall (P: A -> B -> Prop)
           (P_compat: Proper fequiv P)
           (P_dec: forall a b, {P a b} + {~P a b}),
      {forall a b, ~P a b} + {exists a b, P a b}.
  Proof.
    intros P P_compat P_dec.
    pose (P' := fun ab => P (fst ab) (snd ab)).
    assert (P'_compat: Proper fequiv P').
    { intros (a, b) (a', b') [ea eb]; simpl in ea, eb.
      unfold P'; simpl; apply P_compat; auto.
    }
    assert (P'_dec: forall ab, {P' ab} + {~P' ab}).
    { intros (a, b); destruct (P_dec a b); unfold P'; simpl;
        [left|right]; auto.
    }
    destruct (All_In_dec_forall_exists
                all_AB_prop P'_compat P'_dec).
    - left; intros a b.
      specialize (n (a, b)); apply n.
    - right.
      destruct e as [(a, b) h].
      exists a; exists b; apply h.
  Qed.

  Lemma All_In_dec_forall_notforall_2:
    forall (P: A -> B -> Prop)
           (P_dec: forall a b, {P a b} + {~P a b}),
      {Proper fequiv P -> forall a b, P a b} +
      {~(forall a b, P a b)}.
  Proof.
    intros P P_dec.
    pose (P' := fun ab => P (fst ab) (snd ab)).
    assert (P'_dec: forall ab, {P' ab} + {~P' ab}).
    { intros (a, b); destruct (P_dec a b); unfold P'; simpl;
        [left|right]; auto.
    }
    destruct (All_In_dec_forall_notforall
                all_AB_prop _ P'_dec).
    - left; intros P_compat a b.
      assert (P'_compat: Proper fequiv P').
      { intros (aa, bb) (a', b') [ea eb]; simpl in ea, eb.
        unfold P'; simpl; apply P_compat; auto.
      }
      specialize (p P'_compat (a, b)); apply p.
    - right; intros F; contradict n; intros (a, b).
      specialize (F a b); unfold P'; destruct (P_dec a b);
        simpl; auto.
  Qed.

End All_In_2.

Unset Implicit Arguments.
Close Scope signature_scope.
