(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global imports *)
From Coq Require Import RelationPairs.
From Coq Require Import Setoid.
From Coq Require Import ZArith.
From Coq Require Import Bool.
From Coq Require Import BoolEq.
From Coq Require Import Morphisms.
From Coq Require Import SetoidList.
From Coq Require Import SetoidClass.

(** **** Local imports *)
From Padec Require Import SetoidUtil.
From Padec Require Import ListUtils.

(* end show *)
Open Scope Z_scope.
Set Implicit Arguments.

(** * Tools about Fold_right *)
Global Instance fold_right_compat {A: Type} (eqA: relation A)
       {B: Type} (eqB: relation B):
  Proper ((eqB ==> eqA ==> eqA) ==> eqA ==> eqlistA eqB ==> eqA)
         (fold_right (A := A) (B := B)).
Proof.
  intros f1 f2 fa x y eqa l1 l2 eql.
  induction eql; simpl.
  assumption.
  apply fa;assumption.
Qed.

Section Fold_Right_Compat.

  Context {A: Type} `{SA:Setoid A}.
  Context {B: Type} `{SB:Setoid B}.
  Context (f: A -> B -> B).
  Context (f_compat: Proper fequiv f).
  Context (commutative:
             forall x y acc, f x (f y acc) == f y (f x acc)).
  Context (idempotent:
             forall x acc, (f x (f x acc)) == (f x acc)).

  Lemma fold_right_before_after: forall x x0 l,
      (fold_right f (f x x0) l) == (f x (fold_right f x0 l)).
  Proof.
    intros x x0 l.
    induction l.
    simpl.
    reflexivity.
    simpl.
    rewrite IHl.
    rewrite (commutative a x).
    reflexivity.
  Qed.

  Lemma fold_right_left: forall x l,
      (fold_right f x l) == (fold_left (fun b a => f a b) l x).
  Proof.
    intros x l.
    revert x.
    induction l; intros x.
    simpl.
    reflexivity.
    simpl.
    rewrite <- fold_right_before_after.
    apply IHl.
  Qed.

  Lemma fold_right_cons_app:
    forall x0 x l,
      fold_right f x0 (l ++ (x :: nil)) ==
      fold_right f x0 ((x :: nil) ++ l).
  Proof.
    intros x0 x l.
    induction l.
    simpl.
    reflexivity.
    simpl.
    rewrite IHl.
    simpl.
    rewrite (commutative a x).
    reflexivity.
  Qed.

  Lemma fold_right_app_cons_app:
    forall x0 x a b,
      (fold_right f x0 (a ++ (x :: nil) ++ b)) ==
      (fold_right f x0 ((x :: nil) ++ a ++ b)).
  Proof.
    intros x0 x a b.
    repeat rewrite app_assoc.
    rewrite (fold_right_app f (a ++ x :: nil) b).
    rewrite (fold_right_app f ((x :: nil) ++ a) b).
    apply fold_right_cons_app.
  Qed.

  Lemma fold_right_app_app: forall x a b,
      equiv (fold_right f x (a ++ b)) (fold_right f x (b ++ a)).
  Proof.
    intros x a b.
    induction a as [ | h t Hrec].
    simpl.
    rewrite app_nil_r.
    reflexivity.
    transitivity (fold_right f x ((h :: nil) ++ b ++ t)).
    simpl.
    rewrite Hrec.
    reflexivity.
    assert (tmp: h :: t = (h :: nil) ++ t).
    simpl.
    reflexivity.
    rewrite tmp.
    clear tmp.
    symmetry.
    apply (fold_right_app_cons_app x h b t).
  Qed.

  Context {A_dec: Decider (equiv (A:=A))}.

  Lemma InA_cases: forall (x:A) h t,
      InA equiv x (h :: t) ->
      {x == h /\ ~ (InA equiv x t)}
      + {x == h /\ InA equiv x t}
      + {~ (x == h) /\ InA equiv x t}.
  Proof.
    intros x h t inl.
    destruct (A_dec x h) as [eq | neq].
    left.
    destruct (InA_dec A_dec x t) as [int | nint].
    right.
    split; assumption.
    left.
    split; assumption.
    right.
    rewrite InA_cons in inl.
    destruct inl as [inh | int]; split; tauto.
  Qed.

  Lemma no_remove:
    forall x l,
      ~ InA (equiv (A:=A)) x l ->
      eqlistA equiv l (removeA (eqA:=equiv) A_dec x l).
  Proof.
    intros x l;induction l as [|a l hl];simpl removeA.
    left.
    intro not_here;rewrite InA_cons in not_here.
    destruct (A_dec x a) as [hh|hh].
    tauto.
    right.
    reflexivity.
    apply hl;tauto.
  Qed.

  Lemma fold_right_remove:
    forall x acc l,
      InA equiv x l ->
      fold_right f acc l ==
      f x (fold_right f acc (removeA A_dec x l)).
  Proof.
    intros x acc l;induction l as [|a l hl];simpl removeA.
    rewrite InA_nil;tauto.
    intro H;
      destruct (InA_cases H) as [[[h h'] |[h h']] |[h h'] ];
      clear H;
      destruct (A_dec x a) as [hh|hh];
      try tauto;clear hh;simpl fold_right.
    apply f_compat.
    symmetry;auto.
    apply (fold_right_eqlistA _ _ f_compat).
    apply no_remove;auto.
    rewrite <- (idempotent x).
    apply f_compat.
    symmetry;auto.
    apply hl.
    apply h'.
    rewrite commutative.
    apply f_compat.
    reflexivity.
    apply hl.
    apply h'.
  Qed.

  Instance fold_right_compat_equivlist:
    Proper (equiv (A:=B) ==>
                  equivlistA (equiv (A:=A)) ==> equiv (A:=B))
           (fold_right f).
  Proof.
    intros a1 a2 ea l1.
    induction l1 as [|x l hl];simpl fold_right.
    {
      intros [|x l].
      intros _;apply ea.
      intro h;specialize (h x).
      destruct h as [_ h].
      rewrite InA_cons,InA_nil in h.
      contradict h;left;reflexivity.
    }
    {
      intros l2 h.
      assert (hh:=proj1 (h x)).
      rewrite InA_cons in hh.
      specialize (hh (or_introl _ (Equivalence_Reflexive x))).
      destruct (InA_dec A_dec x l).
      {
        transitivity (fold_right f a1 l).
        2:apply hl.
        clear hl.
        destruct (InA_app_app _ i) as [a [b hab]].
        rewrite <- hab.
        rewrite fold_right_app_app.
        simpl fold_right.
        apply idempotent.
        intros a;specialize (h a);
          rewrite InA_cons in h;intuition.
        rewrite H0;auto.
      }
      {
        rewrite (fold_right_remove _ hh).
        apply f_compat.
        reflexivity.
        apply hl.
        apply removeA_equivlistA;auto.
        apply setoid_equiv.
      }
    }
  Qed.

  (** ** equivlistA  / removeA *)
  Lemma remove_duplicates:
    forall x l1 l2,
      InA equiv x l1 ->
      equivlistA equiv l1 l2 ->
      equivlistA equiv (x :: (removeA A_dec x l1)) l2.
  Proof.
    intros x l1 l2 ina eql.
    unfold equivlistA.
    intros y.
    rewrite InA_cons.
    rewrite removeA_InA; auto.
    2: apply setoid_equiv.
    rewrite <- eql.
    split.
    intros [eq_ | [in_ neq_]].
    rewrite eq_.
    assumption.
    assumption.
    intros in_.
    destruct (A_dec x y) as [eq_ | neq_].
    left.
    symmetry.
    assumption.
    right.
    split.
    assumption.
    assumption.
  Qed.

End Fold_Right_Compat.

(** * Fold_right Z.min *)
Section Fold_Right_min_compat.

  Global Instance fold_right_min_compat:
    Proper (equiv ==> equivlistA equiv ==> equiv)
           (fold_right Z.min).
  Proof.
    apply fold_right_compat_equivlist.
    4:intros x y;apply Z.eq_dec.
    apply Z.min_compat.
    intros; rewrite Z.min_assoc,Z.min_assoc,(Z.min_comm x y);
      reflexivity.
    intros; rewrite Z.min_assoc,Z.min_id;reflexivity.
  Qed.

End Fold_Right_min_compat.

(** * Fold_right Z.max *)
Section Fold_Right_max_compat.

  Global Instance fold_right_max_compat:
    Proper (equiv ==> equivlistA equiv ==> equiv)
           (fold_right Z.max).
  Proof.
    apply fold_right_compat_equivlist.
    4:intros x y;apply Z.eq_dec.
    apply Z.max_compat.
    intros; rewrite Z.max_assoc,Z.max_assoc,
            (Z.max_comm x y);reflexivity.
    intros; rewrite Z.max_assoc,Z.max_id;reflexivity.
  Qed.

End Fold_Right_max_compat.

(***********************)
(** * Fold_Right_f_dec *)
(***********************)
Section Fold_Right_f_dec.

  Context {A: Type} `{SA: Setoid A}
          {f: A -> A -> A} {f_compat: Proper fequiv f}
          (f_dec: forall x y:A, {(f x y) == x} + {(f x y) == y}).

  Lemma fold_right_f_dec:
    forall x l,
      {fold_right f x l == x} + {InA equiv (fold_right f x l) l}.
  Proof.
    intros x l.
    induction l as [ | h t Hrec];
      simpl.
    left; reflexivity.
    destruct (f_dec h (fold_right f x t)) as [eql | eqr].
    right; rewrite eql; apply InA_cons_hd; reflexivity.
    destruct Hrec as [inl | inr].
    left; rewrite eqr; assumption.
    right; rewrite eqr; apply InA_cons_tl; assumption.
  Qed.

  Lemma fold_right_f_dec_restr:
    forall x l,
      (forall y, InA equiv y l -> equiv (f y x) y) ->
      {l = nil} + {InA equiv (fold_right f x l) l}.
  Proof.
    intros x l fa.
    induction l as [ | h t Hrec].
    left; reflexivity.
    right.
    simpl.
    destruct Hrec as [eqnil | ina].
    intros y ina'.
    apply fa.
    apply InA_cons_tl.
    assumption.
    subst.
    simpl.
    apply InA_cons_hd.
    apply fa.
    apply InA_cons_hd.
    reflexivity.
    destruct (f_dec h (fold_right f x t)) as [eqa' | ina'].
    rewrite eqa'.
    apply InA_cons_hd.
    reflexivity.
    rewrite ina'.
    apply InA_cons_tl.
    assumption.
  Qed.

End Fold_Right_f_dec.

(** Fold right Z.min decidability *)
Section Fold_Right_Zmin_dec.

  Lemma fold_right_Zmin_dec: forall x l,
    {fold_right Z.min x l == x} +
    {InA equiv (fold_right Z.min x l) l}.
  Proof.
    apply fold_right_f_dec.
    apply Z.min_dec.
  Qed.

  Lemma fold_right_Zmin_dec_restr: forall x l,
      (forall y, InA equiv y l -> Z.min y x == y) ->
      {l = nil} + {InA equiv (fold_right Z.min x l) l}.
  Proof.
    apply fold_right_f_dec_restr.
    apply Z.min_dec.
  Qed.

End Fold_Right_Zmin_dec.

(** Fold right Z.max decidability *)
Section Fold_Right_Zmax_dec.

  Lemma fold_right_Zmax_dec: forall x l,
    {fold_right Z.max x l == x} +
    {InA equiv (fold_right Z.max x l) l}.
  Proof.
    apply fold_right_f_dec.
    apply Z.max_dec.
  Qed.

  Lemma fold_right_Zmax_dec_restr: forall x l,
      (forall y, InA equiv y l -> Z.max y x == y) ->
      {l = nil} + {InA equiv (fold_right Z.max x l) l}.
  Proof.
    apply fold_right_f_dec_restr.
    apply Z.max_dec.
  Qed.

End Fold_Right_Zmax_dec.

(**********************)
(** * Fold_Right_f_le *)
(**********************)

Require Import RelationA.
(* DO NOT MOVE otherwise, a rewrite above doesn't finish *)
Section Fold_Right_f_leA.

  Context
    {A: Type} `{SA: Setoid A}
    {f: A -> A -> A} {f_compat: Proper fequiv f}
    (leA: relation A)
    {leA_compat: Proper fequiv leA}
    {leA_refl: Reflexive leA}
    {leA_trans: Transitive leA}
    (f_leA_left: forall x y, leA (f x y) x)
    (f_leA_right: forall x y, leA (f x y) y)
  .

  Lemma fold_right_f_eqA_leA:
    forall x y l,
      equiv x y ->
      leA (fold_right f x l) y.
  Proof.
    intros x y l eq.
    induction l as [ | h t Hrec].
    simpl.
    rewrite eq.
    reflexivity.
    simpl.
    transitivity (fold_right f x t).
    apply f_leA_right.
    assumption.
  Qed.

  Lemma fold_right_f_InA_leA:
    forall x y l,
      InA equiv y l ->
      leA (fold_right f x l) y.
  Proof.
    intros x y l ina.
    induction l as [ | h t Hrec].
    rewrite InA_nil in ina.
    elim ina.
    simpl.
    rewrite InA_cons in ina.
    destruct ina as [inh | int].
    rewrite <- inh.
    apply f_leA_left.
    transitivity (fold_right f x t).
    apply f_leA_right.
    apply Hrec.
    assumption.
  Qed.

End Fold_Right_f_leA.

(** ** fold right Z.min leA *)

Section Fold_Right_Zmin_leA.

  Lemma fold_right_Zmin_eqA_leA:
    forall x y l,
      x == y ->
      Z.le (fold_right Z.min x l) y.
  Proof.
    apply fold_right_f_eqA_leA.
    apply Z.le_wd.
    unfold Reflexive; apply Z.le_refl.
    unfold Transitive; apply Z.le_trans.
    apply Z.le_min_r.
  Qed.

  Lemma fold_right_Zmin_InA_leA:
    forall x y l,
      InA equiv y l ->
      Z.le (fold_right Z.min x l) y.
  Proof.
    apply fold_right_f_InA_leA.
    apply Z.le_wd.
    intros x y z;apply Z.le_trans.
    apply Z.le_min_l.
    apply Z.le_min_r.
  Qed.

  Lemma fold_right_Zmin_InA:
    forall x l,
      let y := fold_right Z.min x l in
      y == x \/ InA equiv y l.
  Proof.
    induction l  as [|z l hl];simpl.
    left;reflexivity.
    destruct (Zmin_spec z (fold_right Z.min x l))
      as [[_ h]|[_ h]];rewrite h.
    right;left;reflexivity.
    destruct hl as [hl |hl].
    left;assumption.
    right;right;assumption.
  Qed.

End Fold_Right_Zmin_leA.

(** ** fold right Z.max leA *)
Section Fold_Right_Zmax_leA.

  Lemma Zge_wd: Proper (eq ==> eq ==> iff) Z.ge.
  Proof.
    unfold Proper.
    unfold respectful.
    intros x1 x2 eqx y1 y2 eqy.
    subst.
    reflexivity.
  Qed.

  Lemma Zge_refl: Reflexive Z.ge.
  Proof.
    unfold Reflexive.
    intros x.
    rewrite Z.ge_le_iff.
    apply Z.le_refl.
  Qed.

  Lemma Zge_trans: Transitive Z.ge.
  Proof.
    unfold Transitive.
    intros x y z xy yz.
    rewrite Z.ge_le_iff in *.
    transitivity y; assumption.
  Qed.

  Lemma Zge_max_l: forall n m: Z, Z.max n m >= n.
  Proof.
    intros n m.
    rewrite Z.ge_le_iff.
    apply Z.le_max_l.
  Qed.

  Lemma Zge_max_r: forall n m: Z, Z.max n m >= m.
  Proof.
    intros n m.
    rewrite Z.ge_le_iff.
    apply Z.le_max_r.
  Qed.

  Lemma fold_right_Zmax_eqA_leA:
    forall x y l,
      equiv x y ->
      Z.ge (fold_right Z.max x l) y.
  Proof.
    apply fold_right_f_eqA_leA.
    apply Zge_wd.
    apply Zge_refl.
    apply Zge_trans.
    apply Zge_max_r.
  Qed.

  Lemma fold_right_Zmax_InA_leA:
    forall x y l,
      InA equiv y l ->
      Z.ge (fold_right Z.max x l) y.
  Proof.
    apply fold_right_f_InA_leA.
    apply Zge_wd.
    apply Zge_trans.
    apply Zge_max_l.
    apply Zge_max_r.
  Qed.

  Lemma fold_right_Zmax_InA: forall x l,
      let y := fold_right Z.max x l in
      equiv y x \/ InA equiv y l.
    induction l  as [|z l hl];simpl.
    left;reflexivity.
    destruct (Zmax_spec z (fold_right Z.max x l)) as [[_ h]|[_ h]];rewrite h.
    right;left;reflexivity.
    destruct hl as [hl |hl].
    left;assumption.
    right;right;assumption.
  Qed.

End Fold_Right_Zmax_leA.

(** * Fold right min *)
Section Fold_Right_Min.


  Context {A: Type} `{SA:Setoid A}
          {A_dec: Decider (equiv (A:=A))}.

  Variable (lt: relation A)
           (lt_dec: forall a b, { lt a b } + { ~ lt a b })
           (lt_compat: Proper fequiv lt)
           (lt_trans: forall x y z, lt x y -> lt y z -> lt x z)
           (lt_strict: forall x y, lt x y -> ~x == y)
           (lt_total: forall x y, x == y \/ lt x y \/ lt y x)
  .

  Notation le x y := (x == y \/ lt x y).

  Lemma le_refl: forall x, le x x.
  Proof.
    intuition.
  Qed.

  Lemma le_trans: forall x y z, le x y -> le y z -> le x z.
  Proof.
    intros.
    destruct H; destruct H0.
    rewrite H; rewrite H0; left; reflexivity.
    rewrite H; right; apply H0.
    rewrite <- H0; right; apply H.
    right; apply lt_trans with y; assumption.
  Qed.

  Lemma le_antisym: forall x y, le x y -> le y x -> x == y.
  Proof.
    intros.
    destruct H; destruct H0.
    symmetry; assumption.
    assumption.
    symmetry; assumption.
    specialize (lt_trans H H0).
    specialize (lt_strict lt_trans).
    contradict lt_strict; reflexivity.
  Qed.

  Lemma not_lt_le: forall x y, ~lt x y -> le y x.
  Proof.
    intros x y.
    destruct (lt_total x y); [left; symmetry; auto|].
    destruct H; [tauto|right]; auto.
  Qed.

  Lemma not_le_lt: forall x y, ~le x y -> lt y x.
  Proof.
    intros x y h; intuition.
    destruct (lt_total x y); [tauto|].
    destruct H1; [tauto|auto].
  Qed.

  Definition min a b := (if lt_dec a b then a else b).

  Lemma min_compat: Proper fequiv min.
  Proof.
    intros a a' eqa b b' eqb.
    unfold min.
    case (lt_dec a b); case (lt_dec a' b'); intros.
    assumption.
    contradict n;
      rewrite <-(lt_compat _ _ eqa _ _ eqb); assumption.
    contradict n;
      rewrite (lt_compat _ _ eqa _ _ eqb); assumption.
    assumption.
  Qed.

  Lemma min_id: forall a, a == (min a a).
  Proof.
    intros. unfold min.
    case (lt_dec a a); intros; reflexivity.
  Qed.

  Lemma lt_nonsym: forall x y, lt x y -> lt y x -> False.
  Proof.
    intros x y h1 h2;specialize (lt_trans h1 h2).
    specialize (lt_strict lt_trans).
    apply lt_strict;reflexivity.
  Qed.

  Lemma min_commut: forall a b, (min a b) == (min b a).
  Proof.
    intros.
    unfold min.
    case (lt_dec a b); case (lt_dec b a); intros.
    specialize (lt_trans l l0).
    specialize (lt_strict lt_trans).
    contradict lt_strict; reflexivity.
    reflexivity.
    reflexivity.
    specialize (not_lt_le n); intros ne.
    destruct ne.
    symmetry; assumption.
    contradict n0; assumption.
  Qed.

  Lemma min_assoc: forall a b c,
      (min a (min b c)) == (min (min a b) c).
  Proof.
    intros. unfold min.
    case (lt_dec b c); case (lt_dec a b); case (lt_dec a c);
      intros.
    reflexivity.
    specialize (lt_trans l l0); contradict n; assumption.
    case (lt_dec b c); intros.
    reflexivity. contradict n0; assumption.
    case (lt_dec b c); intros.
    reflexivity. contradict n1; assumption.
    reflexivity.     reflexivity.
    case (lt_dec b c); intros. contradict n0; assumption.
    contradict n1.
    specialize (not_lt_le n); intros.
    destruct H.
    rewrite H; assumption.
    apply lt_trans with a; assumption.
    case (lt_dec b c); intros. contradict n1; assumption.
    reflexivity.
  Qed.

  Lemma min_commut': forall a b c,
      (min a (min b c)) == (min b (min a c)).
  Proof.
    intros;rewrite min_assoc,min_assoc.
    apply min_compat.
    apply min_commut.
    reflexivity.
  Qed.

  Lemma min_id': forall a b, (min a (min a b)) == (min a b).
  Proof.
    intros;rewrite min_assoc.
    apply min_compat.
    symmetry;apply min_id.
    reflexivity.
  Qed.

  Lemma fold_right_min_InA:
    forall a b l, a == (fold_right min b l) ->
                  InA equiv a (b :: l).
  Proof.
    intros a b l heq. revert a b heq.
    induction l as [| x l IH]; simpl.
    intros a b heq.
    rewrite InA_cons.
    left; assumption.
    intros a b heq.
    unfold min at 1 in heq.
    revert heq; case(lt_dec x (fold_right min b l));
      intros hm heq.
    rewrite InA_cons.
    right; rewrite InA_cons.
    left; assumption.
    rewrite InA_cons.
    case (A_dec a b); intros eqab.
    left; assumption.
    right; rewrite InA_cons.
    specialize (IH _ _ heq).
    rewrite InA_cons in IH.
    destruct IH. contradict eqab; assumption.
    right; assumption.
  Qed.

  Lemma fold_right_min_after_before:
    forall a b l,
      (fold_right min a (b :: l)) == (fold_right min b (a :: l)).
  Proof.
    intros a b l.
    induction l; simpl.
    apply min_commut.
    simpl in IHl.
    rewrite (min_compat _ _ (Equivalence_Reflexive b) _ _
                        (min_commut a0 (fold_right min a l))).
    rewrite min_assoc.
    rewrite (min_compat _ _ (Equivalence_Reflexive a) _ _
                        (min_commut a0 (fold_right min b l))).
    rewrite min_assoc.
    apply (min_compat _ _ IHl _  _(Equivalence_Reflexive a0) ).
  Qed.

  Lemma fr_min_le: forall a l, le (fold_right min a l) a.
  Proof.
    induction l.
    simpl.
    apply le_refl.
    simpl.
    unfold min at 1 3.
    destruct IHl;
      case (lt_dec a0 (fold_right min a l)); intros.
    rewrite (lt_compat _ _ (Equivalence_Reflexive a0) _ _ H) in
        l0; right; assumption.
    rewrite H; left; reflexivity.
    right; apply lt_trans with (fold_right min a l).
    assumption.
    assumption.
    right; assumption.
  Qed.

  Lemma fr_min_tool:
    forall a l, (fold_right min a (a :: l)) ==
                (fold_right min a l).
  Proof.
    intros a l.
    simpl.
    unfold min at 1.
    generalize (fr_min_le a l).
    case (lt_dec a (fold_right min a l)); intros.
    destruct H.
    symmetry; assumption.
    specialize (lt_trans l0 H).
    specialize (lt_strict lt_trans).
    contradict lt_strict; reflexivity.
    reflexivity.
  Qed.

  Lemma fr_min_tool':
    forall a b l,
      InA equiv a (b::l) -> (fold_right min a (b :: l)) ==
                            (fold_right min b l).
  Proof.
    induction l as [| h l IH]; intros H.
    rewrite InA_cons in H.
    destruct H as [H|H].
    rewrite (min_compat _ _ (Equivalence_Reflexive b) _ _ H).
    unfold min.
    case (lt_dec b b); reflexivity.
    rewrite InA_nil in H; elim H.

    case (A_dec a h); intros EA.
    rewrite fold_right_min_after_before.
    assert (X: equivlistA equiv (a::h::l) (h::l)).
    intros x.
    rewrite EA.
    repeat rewrite InA_cons.
    intuition.
    apply fold_right_compat_equivlist; try assumption.
    apply min_compat.
    apply min_commut'.
    apply min_id'.
    reflexivity.

    assert (X: InA equiv a (b::l)).
    rewrite InA_cons in H.
    rewrite InA_cons.
    destruct H; [left;assumption|].
    rewrite InA_cons in H.
    destruct H; [contradict EA|right]; assumption.
    specialize (IH X).

    assert
      (ADM:
         forall a a' l l',
           (fold_right min a l) == (fold_right min a' l') ->
           forall b,
             (fold_right min a (b::l)) ==
             (fold_right min a' (b::l'))).
    {
      intros.
      simpl.
      apply min_compat.
      reflexivity.
      apply H0.
    }
    rewrite fold_right_compat_equivlist;[apply ADM,IH|..];auto.
    apply min_compat.
    apply min_commut'.
    apply min_id'.
    apply  equivlistA_permute_heads; assumption.
  Qed.

  Lemma removeA_insert:
    forall a b l,
      ~ a == b ->
      equivlistA equiv (b::(removeA A_dec a l))
                 (removeA A_dec a (b::l)).
  Proof.
    destruct l as [| h l]; intros N.
    simpl.
    destruct (A_dec a b);
      [contradict e; assumption| reflexivity].
    simpl.
    destruct (A_dec a h).
    destruct (A_dec a b);
      [contradict e0; assumption| reflexivity].
    destruct (A_dec a b);
      [contradict e; assumption| reflexivity].
  Qed.

  Lemma fold_right_min_cons:
    forall a b la lb,
      equivlistA equiv (a :: la) (b :: lb) ->
      (fold_right min a la) == (fold_right min b lb).
  Proof.
    induction la as [| ha la IH].

    (* la=nil *)
    induction lb as [| hb lb].
    unfold equivlistA; simpl; intros H.
    specialize (H a).
    repeat rewrite InA_cons in H; repeat rewrite InA_nil in H.
    assert (tmp: a == b \/ False).
    apply H; left; reflexivity.
    intuition.

    unfold equivlistA in *; simpl; intros H.
    simpl in IHlb.
    unfold min at 1.
    case (lt_dec hb (fold_right min b lb)); intros _.
    specialize (H hb).
    repeat rewrite InA_cons in H; repeat rewrite InA_nil in H.
    symmetry.
    assert (tmp: hb == a \/ False).
    apply H.
    right. left.
    reflexivity.
    destruct tmp as [tmp | tmp]; [assumption|elim tmp].

    apply IHlb.
    intros x.
    repeat rewrite InA_cons; repeat rewrite InA_nil.
    generalize (H x); intros Hx.
    repeat rewrite InA_cons in Hx; repeat rewrite InA_nil in Hx.
    rewrite Hx.
    clear Hx.
    generalize (H hb); intros Hhb.
    repeat rewrite InA_cons in Hhb;
      repeat rewrite InA_nil in Hhb.
    assert (tmp: hb == a \/ False).
    apply Hhb.
    right. left.
    reflexivity.
    assert (eqhb: hb == a).
    destruct tmp as [tmp | tmp]; [assumption|elim tmp].
    rewrite eqhb.
    clear tmp Hhb eqhb.
    generalize (H b); intros Hb.
    repeat rewrite InA_cons in Hb; repeat rewrite InA_nil in Hb.
    assert (tmp: b == a \/ False).
    apply Hb.
    left.
    reflexivity.
    assert (eqb: b == a).
    destruct tmp as [tmp | tmp]; [assumption|elim tmp].
    rewrite eqb.
    clear tmp Hb eqb.
    intuition.

    (* la <> nil *)
    intros lb H.
    case (InA_dec A_dec ha (a::la)); intro inha.
    generalize (H ha); intros H'.
    destruct H' as [H' _].
    rewrite InA_cons in H'.
    rewrite fold_right_min_after_before.
    rewrite (fr_min_tool' inha).
    apply IH.
    rewrite <- H.
    intros x.
    repeat rewrite InA_cons.
    rewrite InA_cons in inha.
    destruct inha as [inha|inha].
    rewrite inha. intuition.
    split. intuition.
    intros [HY | [HY|HY]].
    left; assumption.
    right; rewrite HY; assumption.
    right; assumption.

    case (A_dec ha b); intros EB.

    rewrite fold_right_min_after_before.
    assert (OK: (fold_right min b (b :: lb)) ==
                (fold_right min ha (ha :: a :: la))).
    apply fold_right_compat_equivlist; try assumption.
    apply min_compat.
    apply min_commut'.
    apply min_id'.
    symmetry; assumption.
    symmetry.
    rewrite equivlistA_permute_heads; assumption.

    rewrite fr_min_tool in OK.
    symmetry; rewrite fr_min_tool in OK.
    assumption.

    remember (removeA A_dec ha lb) as l'b.
    assert (XX: equivlistA equiv (a::la) (b::l'b)).
    rewrite Heql'b.
    rewrite removeA_insert.

    rewrite <- removeA_equivlistA; try assumption.
    reflexivity.
    apply setoid_equiv.
    apply inha.
    rewrite equivlistA_permute_heads; assumption.
    assumption.
    specialize (IH _ XX).
    simpl.
    rewrite min_compat.
    2: reflexivity.
    2: apply IH.
    assert (v: (min ha (fold_right min b l'b)) ==
               (fold_right min b (ha::l'b))).
    simpl; reflexivity.
    rewrite v; clear v.
    apply fold_right_compat_equivlist; try assumption.
    apply min_compat.
    apply min_commut'.
    apply min_id'.
    reflexivity.
    rewrite Heql'b; apply remove_duplicates; try assumption.
    2: reflexivity.
    specialize (H ha).
    destruct H as [H _].
    rewrite InA_cons in H.
    rewrite InA_cons in H.
    assert (tmp:  InA equiv ha (b :: lb)).
    apply H.
    right. left; reflexivity.
    rewrite InA_cons in tmp.
    destruct tmp; [contradict EB|]; assumption.
  Qed.

End Fold_Right_Min.

Close Scope Z_scope.
Unset Implicit Arguments.
