(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Feb 2016
 *)
(* begin show *)

(** **** Global imports *)
Require Import Relations.
Require Import SetoidClass.

(** **** Local imports *)
Require Import SetoidUtil.

(* end show *)
Open Scope signature_scope.
Set Implicit Arguments.


(** ** Setoid and Partial Setoid on options *)

(** equality on option *)
Definition eqoptionA {A: Type} (eqA: relation A): relation (option A) :=
  fun oa1 oa2 => match oa1,oa2 with 
                 | None, None => True
                 | Some a1, Some a2 => eqA a1 a2
                 | _, _ => False
                 end.

(** comparison on option *)
Definition leoptionA {A: Type} (eqA: relation A): relation (option A) :=
  fun oa1 oa2 => match oa1, oa2 with 
                 | None, _ => True
                 | Some a1, Some a2 => eqA a1 a2
                 | _, _ => False
                 end.

Global Instance leoptionA_compat {A: Type} (eqA: relation A) `{P: PER _ eqA}:
  Proper (eqoptionA eqA ==> eqoptionA eqA ==> iff) (leoptionA eqA).
Proof.
  intros oa1 oa2 ea ob1 ob2 eb. 
  destruct oa1 as [ a1 |]; destruct oa2 as [ a2|]; 
    destruct ob1 as [ b1|]; destruct ob2 as [b2|]; 
      simpl in *; try tauto.
  destruct P as [S T]; unfold Symmetric in S; unfold Transitive in T.
  split; intros X.
  apply T with b1; [| assumption].
  apply T with a1; [ apply S|]; assumption.
  apply T with b2; [| apply S; assumption].
  apply T with a2; assumption.
Qed.

Global Instance leoptionA_transitive {A: Type} (eqA: relation A):
  Transitive eqA -> Transitive (leoptionA eqA).
Proof.
  intros  H.
  intros [a1|] [a2|] [a3|];simpl;try tauto.
  apply H.
Qed.

(** PartialSetoid/Setoid for options based on a
    PartialSetoid/Setoid *)
Global Instance eqoptionA_per {A: Type} (eqA: relation A) `{P: PER _ eqA}:
  PER (eqoptionA eqA).
Proof.
  split.
  intros [x|] [y|]; simpl; trivial; apply P.
  intros [x|] [y|] [z|]; simpl; trivial; [apply P|tauto].
Qed.

Global Instance eqoptionA_equiv {A: Type} (eqA: relation A)
         `{P: Equivalence _ eqA}: Equivalence (eqoptionA eqA).
Proof.
  split.
  intros [x|]; simpl; auto; apply P.
  intros [x|] [y|]; simpl; trivial; apply P.
  intros [x|] [y|] [z|]; simpl; trivial; [apply P|tauto].
Qed.

Global Instance option_PartialSetoid {A} `{SA: PartialSetoid A}:
  PartialSetoid (option A) := {| pequiv := eqoptionA pequiv |}.

Global Instance option_Setoid {A} `{SA: Setoid A}:
  Setoid (option A) := {| equiv := eqoptionA equiv |}.


(** ** Decidability *)
Lemma eqoptionA_dec {A:Type} (eqA: relation A):
  Decider eqA -> Decider (eqoptionA eqA).
Proof.
  intros eqA_dec x y; destruct x; destruct y; simpl; auto. 
Qed.

Lemma leoptionA_dec {A:Type} (eqA: relation A):
  Decider eqA -> Decider (leoptionA eqA).
Proof.
  intros eqA_dec x y; destruct x; destruct y; auto; unfold leoptionA.
  apply eqA_dec.
  left; trivial.
  left; trivial.
Qed.

(** *** Properties assuming Setoids *)
Section with_Setoid.

  Context {A} `{A_Setoid: Setoid A}.

  Instance leoptionA_preorder: PreOrder (leoptionA equiv).
  Proof.
    split.
    intros [a|];simpl.
    reflexivity.
    trivial.
    intros [a1|] [a2|] [a3|];simpl;try tauto.
    apply Equivalence_Transitive.
  Qed.

  Instance leoptionA_partialorder: PartialOrder equiv (leoptionA equiv).
  Proof.
    intros [a1|] [a2|];cbv;intuition.
  Qed.
  
  Global Instance Some_compat : Proper (equiv ==> equiv) (@Some A).
  Proof.
    intros x x' Ex;simpl;auto.
  Qed.

  Global Instance None_compat : Proper equiv (@None A) := I.
  (* I est la preuve de True. *)

  (** **** *)
  Context {B: Type} `{B_Setoid: Setoid B}.

  Global Instance option_rect_compat:
    Proper ((equiv (A := A) ==> equiv (A := B))
              ==> equiv ==> equiv ==> equiv) (option_rect (fun _=>B)).
  Proof.
    intros f1 f2 ef b1 b2 eb [a1|] [a2|] h;simpl in h|-*;auto || elim h.
  Qed.

  Global Instance option_map_compat:
    Proper ((equiv (A := A) ==> equiv (A := B))
              ==> equiv ==> equiv) (option_map (B := B)).
  Proof.
    intros f1 f2 ef [a1|] [a2|] ea; simpl in *; try tauto; auto.
  Qed.

End with_Setoid.

(** *** Properties assuming Partial Setoids *)
Section with_PartialSetoid.

  Context {A:Type} `{A_PartialSetoid: PartialSetoid A}.

  Global Instance Some_compat': Proper (pequiv ==> pequiv) (@Some A).
  Proof.
    intros x x' Ex;simpl;auto.
  Qed.

  Global Instance None_compat': Proper pequiv (@None A) := I.
  (* I est la preuve de True. *)

  (** **** *)
  Context {B:Type} `{B_Psetoid: PartialSetoid B}.

  Global Instance option_rect_compat':
    Proper ((pequiv (A := A) ==> pequiv (A := B))
              ==> pequiv ==> pequiv ==> pequiv) (option_rect (fun _=>B)).
  Proof.
    intros f1 f2 ef b1 b2 eb [a1|] [a2|] h;simpl in h|-*;auto || elim h.
  Qed.

  Global Instance option_map_compat' eqA eqB:
    Proper ((eqA ==> eqB) ==> (eqoptionA eqA)
                          ==> (eqoptionA eqB)) (@option_map A B).
  Proof.
    intros f1 f2 ef [a1|] [a2|] ea; simpl in *; try tauto.
    apply ef.
    apply ea.
  Qed.

End with_PartialSetoid.

Unset Implicit Arguments.
Close Scope signature_scope.
