(**
#<a href="http://www-verimag.imag.fr/~altisen/PADEC/">
A Framework for Certified Self-Stabilization
</br>
PADEC Coq Library
</a>#

- Nov 2016
 *)
(* begin show *)

From Coq Require Import Setoid.
From Coq Require Import SetoidList.
From Coq Require Import Morphisms.
From Coq Require Import RelationClasses.
From Coq Require Import SetoidClass.
From Coq Require Import RelationPairs.

From Padec Require Import SetoidUtil.
From Padec Require Import Stream.

(* end show *)
Open Scope signature_scope.
Set Implicit Arguments.


Section Relational_Stream.

  Context {A: Type} `{SA: PartialSetoid A}.
  
  (*
    CoInductive Relational_Stream (R: relation A) : Stream-> Prop:=
    | R_stream_one  : forall a:A, (forall x, ~R x a)->Relational_Stream R (s_one a)
    | R_stream_cons : forall (a:A)(S:Stream), R (s_hd S) a->Relational_Stream R S->
                                                Relational_Stream R (s_cons a S).*)

  Notation continue_R R :=
    (fun s => match s with | s_one _ => False
                      | s_cons a s => R (s_hd s) a end).
  Notation stop_P P :=
    (fun s => match s with | s_one a => P a | s_cons _ _ => False end).

  Definition Relational_Stream (R: relation A) (P: A -> Prop): Stream -> Prop :=
    WUntil (continue_R R) (stop_P P).

  Lemma Relational_Stream_inv_1:
    forall R P a s, Relational_Stream R P (s_cons a s) -> R (s_hd s) a.
  Proof.
    intros R P a s H; inversion H; subst; auto; tauto.
  Qed.

  Lemma Relational_Stream_inv_2:
    forall R P a s, Relational_Stream R P (s_cons a s) -> Relational_Stream R P s.
  Proof.
    intros R P a s H; inversion H; subst; auto; tauto.
  Qed.

  Lemma Relational_Stream_inv_0:
    forall R P a, Relational_Stream R P (s_one a) -> P a.
  Proof.
    intros R P a H; inversion H; subst; auto; tauto.
  Qed.

  Lemma Always_Relational_Stream:
    forall R P a, Relational_Stream R P a -> Always (fun s => Relational_Stream R P s) a.
  Proof.
    cofix Always_Relational_Stream.
    intros R P a H;destruct a.
    apply Relational_Stream_inv_0 in H.
    constructor 1; constructor 1;auto.
    constructor 2.
    constructor 2.
    apply Relational_Stream_inv_1 in H;auto.
    apply Relational_Stream_inv_2 in H;auto.
    apply Always_Relational_Stream;apply Relational_Stream_inv_2 in H;auto.
  Qed.

  (** The restriction of a relation on its right member to a property P*)
  Definition Restriction (R: relation A) (P: A -> Prop) := (fun x y => R x y /\ P y).

  (** The restriction of R to P is included in R *)
  Lemma subrelation_Restriction: forall R P, subrelation (Restriction R P) R.
  Proof.
    intros;unfold Restriction;unfold subrelation;intuition.
  Qed.


  Global Instance Relational_Stream_compat:
    Proper (fequiv ==> fequiv ==> pequiv ==> impl) Relational_Stream.
  Proof.
    cofix h.
    intros R R' eqR P P' eqP [a | a s] [a' | a' s'] es hh;
      inversion es; subst.
    - left; apply Relational_Stream_inv_0 in hh; revert hh.
      apply eqP; auto.
    - right.
      apply Relational_Stream_inv_1 in hh; revert hh.
      apply eqR; [apply hd_compat |]; auto.
      apply (h _ _ eqR _ _ eqP _ _ H4).
      apply Relational_Stream_inv_2 in hh; auto.
  Qed.

  Global Instance Relational_Stream_compat_iff: Proper fequiv Relational_Stream.
  Proof.
    intros R R' eqR P P' eqP s s' es.
    split; apply Relational_Stream_compat; auto.
    { intros a1 a2 ea b1 b2 eb; symmetry; apply eqR; symmetry; auto. }
    { intros a1 a2 ea; symmetry; apply eqP; symmetry; auto. }
    { symmetry; auto. }
  Qed.

  (** If R preserves P and the first element of a stream respects P then it is the same
        thing to be a stream carrying information about R and be a stream carrying
        information about the restriction of R.
   *)
  Lemma Restriction_intro: forall (R: relation A)(Z P: A -> Prop),
      (forall x y, P x -> R y x -> P y) ->
      forall e, P (s_hd e) ->
                Relational_Stream R Z e <->
                Relational_Stream (Restriction R P) Z e.
  Proof.
    intros R Z P P_compat e P_e.
    split; revert e P_e;cofix Restriction_intro;
      intros e P_e;destruct e.
    { constructor 1;inversion H; auto. }
    { constructor 2.
      split;auto.
      inversion H; subst; [tauto|auto].
      apply Restriction_intro.
      apply P_compat with (x:=a);auto.
      apply Relational_Stream_inv_1 with Z; auto.
      apply Relational_Stream_inv_2 with a; auto.
    }
    { intros H; constructor 1; inversion H; auto. }
    { intros H; constructor 2.
      inversion H; subst; [tauto|apply H2].
      apply Restriction_intro.
      apply P_compat with a; auto.
      inversion H; subst; [tauto| apply H2].
      apply Relational_Stream_inv_2 with a; auto.
    }
  Qed.

  (** Eventually when the implication is conditionned on a third property P stable by R.
   *)
  Lemma Eventually_impl2: forall (R: relation A) (Z: A -> Prop) (P: A -> Prop) (B C: Stream -> Prop),
      (forall x y, P x -> R y x -> P y) ->
      (forall e, P (s_hd e) -> B e -> C e) ->
      forall e, P (s_hd e) ->
                Relational_Stream R Z e ->
                Eventually B e ->
                Eventually C e.
  Proof.
    intros R Z P C D P_compat P_C_D e P_s_hd H H0.
    induction H0.
    apply eventually_now;auto.
    apply eventually_later.
    apply IHEventually.
    apply P_compat with (x:=a);auto.
    apply Relational_Stream_inv_1 with Z; auto.
    apply Relational_Stream_inv_2 with a; auto.
  Qed.


  (** Let R be a relation and P,Q,Z three properties on A.
     if
     #<ul>
     <li>P is preserved by R</li>
     <li>for all stream e that are relational Streams, P x implies Eventually Q</li>
     </ul>
     then forall all stream carrying information about R such that the first element respects P
     eventually Q.
   *)

  Lemma Eventually_forall_always_enventually:
    forall (R:relation A) (Z P:A->Prop) (Q:Stream->Prop),
      (forall x y,P x-> R y x ->P y)->
      (forall e, P (s_hd e)-> Relational_Stream R Z e->Eventually Q e)->
      forall e, Relational_Stream R Z e->
                P (s_hd e) -> Always (Eventually Q ) e.
  Proof.
    intros R Z P Q H.
    cofix Eventually_forall_always_enventually.
    intros H0 e H1 H2.
    destruct e.
    constructor 1.
    apply H0;auto.
    constructor 2.
    apply H0;auto.
    inversion H1; subst;[tauto|].
    assert (P (s_hd e)).
    apply (H a (s_hd e));auto.
    apply Eventually_forall_always_enventually;auto.
  Qed.

  Lemma Relational_Acc_Eventually:
    forall (R:relation A) (Z:A -> Prop) (Q:Stream->Prop),
      (forall e, Z (s_hd e) -> Q e) ->
      forall e, Relational_Stream R Z e -> Acc R (s_hd e) -> Eventually Q e.
  Proof.
    intros R Z Q hZQ e hrel w.
    pose (a:=s_hd e).
    assert (ea:a=s_hd e) by reflexivity.
    fold a in w.
    clearbody a.
    revert e ea hrel.
    induction w as [ eee _ hrecw];intros e ea.
    intro rs;inversion rs;subst.
    destruct e as [a | a e].
    left;auto.
    elim H.
    right;apply hrecw with (s_hd s);auto.
  Qed.

  Section Weak_Prefix.

    Section Weak_Prefix_with_dep.
      Variable (Safe: Stream (A := A) -> Prop)
               (P: Stream (A := A) -> Prop).
      Hypothesis (P_dec: forall s, Safe s -> {P s}+{~P s}).

      CoFixpoint weak_prefix_until_P s (asa: Always Safe s) :=
        match s as ns return Always Safe ns -> Stream with
        | s_one a => fun _ => s_one a
        | s_cons a s' => fun asa' =>
                           if P_dec (Always_P_P asa) then s_one a
                           else s_cons a (weak_prefix_until_P
                                            (Always_inv_2 asa'))
        end asa.

      Lemma weak_prefix_until_P_decompose:
        forall s asa,
          weak_prefix_until_P asa =
          match s as ns return Always Safe ns -> Stream with
          | s_one a => fun _ => s_one a
          | s_cons a s' => fun asa' =>
                             if P_dec (Always_P_P asa) then s_one a
                             else s_cons a (weak_prefix_until_P
                                              (Always_inv_2 asa'))
          end asa.
      Proof.
        intros s asa.
        rewrite Stream_decomposition_lemma with (weak_prefix_until_P asa).
        destruct s as [ a | a s ]; simpl; auto.
        destruct (P_dec (Always_P_P asa)); auto.
      Qed.

      Lemma weak_prefix_until_P_hd: forall s (asa: Always Safe s),
          s_hd (weak_prefix_until_P asa) = (s_hd s).
      Proof.
        intros s asa.
        rewrite weak_prefix_until_P_decompose.
        destruct s as [ a | a s ]; simpl; auto.
        destruct (P_dec (Always_P_P asa)); auto.
      Qed.

    End Weak_Prefix_with_dep.

    Lemma finite_weak_prefix:
      forall (P Safe: Stream -> Prop)
             (P_dec: forall s, Safe s -> {P s} + {~P s})
             (Q: A -> Prop) (incl: forall s, Q (s_hd s) -> P s)
             (s: Stream) (asa: Always Safe s),
        Always (fun e => match e with
                    s_one g => Q g | s_cons _ _ => True end) s ->
        finite (weak_prefix_until_P _ P_dec asa) ->
        Eventually (fun suf => P suf) s.
    Proof.
      intros P Safe P_dec Q incl s asa rs fin.
      remember (weak_prefix_until_P _ P_dec asa) as w.
      revert Heqw rs. revert asa. revert s.
      induction fin as [[a | a ss] hss| a s p ih]; try tauto.
      - intros [aa | aa s] asa hw rs.
        + left; apply incl; inversion rs; auto. 
        + rewrite weak_prefix_until_P_decompose in hw.
          destruct (P_dec (s_cons aa s) (Always_P_P asa)).
          * left; auto.
          * inversion hw.
      - intros [aa | aa ss] asa hw rs.
        + left; apply incl; inversion rs; auto.
        + rewrite weak_prefix_until_P_decompose in hw.
          destruct (P_dec (s_cons aa ss) (Always_P_P asa)).
          * inversion hw.
          * right; eapply ih.
            inversion hw; auto.
            inversion rs; auto.
    Qed.


  End Weak_Prefix.

  (** * Finite Relational Stream:
   definitions with Acc and well_founded
   *)
  Section Finite_Relational_Stream.

    Variable (R: relation A) (P: A -> Prop).
    Notation RP := (fun a b => ~P b /\ R a b).

    Lemma Acc_finite_rel_stream:
      forall init,
        Acc RP init ->
        (forall s, init = s_hd s ->
                   Always (fun s => match s with
                                      s_one a => P a
                                    | s_cons a s => RP (s_hd s) a end) s
                   -> finite s).
    Proof.
      intros init h.
      induction h as [a h ih] using Acc_ind_dep.
      intros [b | b s] hd rs; [left; auto | right].
      apply ih with (s_hd s); auto.
      inversion rs; auto.
      simpl in hd; rewrite hd; auto.
      inversion rs; auto.
    Qed.
    (* TODO: converse should be true!!! *)

    Lemma wf_finite_rel_stream:
      well_founded RP ->
      (forall s,
          Always (fun s => match s with
                             s_one a => P a
                           | s_cons a s => RP (s_hd s) a end) s
          -> finite s).
    Proof.
      intros h.
      intros s rs; eapply Acc_finite_rel_stream.
      apply h.
      reflexivity.
      auto.
    Qed.

  End Finite_Relational_Stream.

  Lemma rel_stream_preserved:
    forall R P s,
      Relational_Stream R P s ->
      forall Safe Q Q_dec (asa: Always Safe s),
        Relational_Stream (fun a b => ~Q b /\ R a b)
                          (fun a => P a \/ Q a)
                          (weak_prefix_until_P (fun s => Q (s_hd s)) Q_dec asa).
  Proof.
    intros R P s rs Safe Q Q_dec asa.
    revert s rs Q_dec asa.
    cofix h.
    intros [ a | a s ] rs Q_dec asa;
      rewrite weak_prefix_until_P_decompose;
      simpl.
    - left; left; apply Relational_Stream_inv_0 in rs; auto.
    - destruct (Q_dec (s_cons a s) (Always_P_P asa)).
      + left; right; auto.
      + right.
        rewrite weak_prefix_until_P_hd.
        apply Relational_Stream_inv_1 in rs; split; auto.
        apply h.
        apply Relational_Stream_inv_2 in rs; auto.
  Qed.


End Relational_Stream.

Unset Implicit Arguments.
Close Scope signature_scope.

